import asyncio
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import AsyncIterator

from ..messages import Tick, Message, Zero
from .base import Node
from .registry import NodeFactory
import time


@NodeFactory.register("tick")
class TickNode(Node[Zero, Tick]):
    def __init__(self, redis_url: str = "redis://localhost:6379/0"):
        super().__init__(
            input_channel_types=[],
            output_channel_types=[
                ("tick/millis/10", Tick),
                ("tick/millis/20", Tick),
                ("tick/millis/33", Tick),
                ("tick/millis/50", Tick),
                ("tick/millis/100", Tick),
                ("tick/secs/1", Tick),
            ],
            redis_url=redis_url,
        )

    async def tick_at_given_interval(self, channel: str, interval: float) -> None:
        tick_count = 0
        last: float | None = None
        last_sleep = interval
        while True:
            await self.r.publish(
                channel, Message[Tick](data=Tick(tick=tick_count)).model_dump_json()
            )
            tick_count += 1
            now = time.time()
            if last is not None:
                last_sleep = last_sleep - (now - last - interval)
            await asyncio.sleep(last_sleep)
            last = now

    async def event_loop(self) -> None:
        await asyncio.gather(
            self.tick_at_given_interval("tick/millis/10", 0.01),
            self.tick_at_given_interval("tick/millis/20", 0.02),
            self.tick_at_given_interval("tick/millis/33", 0.033),
            self.tick_at_given_interval("tick/millis/50", 0.05),
            self.tick_at_given_interval("tick/millis/100", 0.1),
            self.tick_at_given_interval("tick/secs/1", 1.0),
        )

    async def __aenter__(self) -> Self:
        return await super().__aenter__()

    async def event_handler(
        self, _: str, __: Message[Zero]
    ) -> AsyncIterator[tuple[str, Message[Tick]]]:
        raise NotImplementedError("TickNode does not have an event handler.")
        yield "", Message[Tick](data=Tick(tick=0))



from datetime import datetime
from typing import Any, Annotated, Generic, TypeVar

from .registry import DataModelFactory
from .base import DataModel
from pydantic import Field, PlainValidator, PlainSerializer, WithJsonSchema, BaseModel


@DataModelFactory.register("zero")
class Zero(DataModel):
    pass


@DataModelFactory.register("tick")
class Tick(DataModel):
    tick: int


@DataModelFactory.register("float")
class Float(DataModel):
    value: float


def hex_bytes_validator(o: Any) -> bytes:
    if isinstance(o, bytes):
        return o
    elif isinstance(o, bytearray):
        return bytes(o)
    elif isinstance(o, str):
        return bytes.fromhex(o)
    raise ValueError(f"Expected bytes, bytearray, or hex string, got {type(o)}")


HexBytes = Annotated[
    bytes,
    PlainValidator(hex_bytes_validator),
    PlainSerializer(lambda b: b.hex()),
    WithJsonSchema({"type": "string"}),
]


@DataModelFactory.register("image")
class Image(DataModel):
    image: HexBytes


@DataModelFactory.register("text")
class Text(DataModel):
    text: str


@DataModelFactory.register("audio")
class Audio(DataModel):
    audio: HexBytes


T = TypeVar("T", bound=DataModel)


class DataEntry(BaseModel, Generic[T]):
    timestamp: datetime = Field(default_factory=datetime.now)
    channel: str
    data: T



import asyncio
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any, AsyncIterator

from ..messages.commons import DataEntry

from .base import Node
from .registry import NodeFactory
from ..messages import DataModel, Zero, Message
from ..messages.registry import DataModelFactory

from aiofiles import stdout
from aiofiles.threadpool.text import AsyncTextIndirectIOWrapper


@NodeFactory.register("print")
class PrintNode(Node[DataModel, Zero]):
    def __init__(
        self,
        print_channel_types: dict[str, str],
        redis_url: str,
    ):
        input_channel_types: list[tuple[str, type[DataModel]]] = []
        for channel, channel_type_string in print_channel_types.items():
            input_channel_types.append(
                (channel, DataModelFactory.registry[channel_type_string])
            )

        super().__init__(
            input_channel_types=input_channel_types,
            output_channel_types=[],
            redis_url=redis_url,
        )
        self.output: AsyncTextIndirectIOWrapper | None = None
        self.write_queue: asyncio.Queue[DataEntry[DataModel]] = asyncio.Queue()
        self.write_task: asyncio.Task[None] | None = None

    async def __aenter__(self) -> Self:
        self.output = stdout
        self.write_task = asyncio.create_task(self.write_to_screen())
        return await super().__aenter__()

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        del self.output
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def write_to_screen(self) -> None:
        while self.output:
            data_entry = await self.write_queue.get()
            await self.output.write(data_entry.model_dump_json() + "\n")
            await self.output.flush()
            self.write_queue.task_done()

    async def event_handler(
        self, input_channel: str, input_message: Message[DataModel]
    ) -> AsyncIterator[tuple[str, Message[Zero]]]:
        if input_channel in self.input_channel_types:
            await self.write_queue.put(
                DataEntry[self.input_channel_types[input_channel]](  # type: ignore[name-defined]
                    channel=input_channel, data=input_message.data
                )
            )
        else:
            yield input_channel, Message[Zero](data=Zero())



import sys
from typing import Any, AsyncIterator, Optional, TYPE_CHECKING
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from .base import Node
from .registry import NodeFactory
from ..messages import Audio, Zero, Message

# For type checking and IDE support
if TYPE_CHECKING:
    import pyaudio

# Runtime import
try:
    import pyaudio
    PYAUDIO_AVAILABLE = True
except ImportError:
    PYAUDIO_AVAILABLE = False

@NodeFactory.register("speaker")
class SpeakerNode(Node[Audio, Zero]):
    def __init__(
        self,
        input_channel: str,
        redis_url: str,
    ):
        if not PYAUDIO_AVAILABLE:
            raise ImportError(
                "PyAudio is not available. Please install it to use the SpeakerNode."
            )
        
        super().__init__(
            input_channel_types=[(input_channel, Audio)],
            output_channel_types=[],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.audio: 'pyaudio.PyAudio' = pyaudio.PyAudio()
        self.stream: Optional['pyaudio.Stream'] = None

    async def __aenter__(self) -> Self:
        if PYAUDIO_AVAILABLE:
            self.stream = self.audio.open( 
                format=pyaudio.paInt16, channels=1, rate=44100, output=True
            )
        return await super().__aenter__()

    async def __aexit__(self, _: Any, __: Any, ___: Any) -> None:
        if self.stream:
            self.stream.stop_stream()  
            self.stream.close() 
        if PYAUDIO_AVAILABLE:
            self.audio.terminate()
        return await super().__aexit__(_, __, ___)

    async def event_handler(
        self, input_channel: str, message: Message[Audio]
    ) -> AsyncIterator[tuple[str, Message[Zero]]]:
        if input_channel == self.input_channel:
            if self.stream:
                self.stream.write(message.data.audio)
            else:
                raise ValueError(
                    "Stream is not initialized. Please use the async context manager."
                )
        else:
            yield "", Message[Zero](data=Zero())


from .base import Message, DataModel
from .commons import Zero, Tick, Image, Float, Audio
from .registry import DataModelFactory

__all__ = [
    "Zero",
    "Message",
    "Tick",
    "Image",
    "Float",
    "DataModelFactory",
    "DataModel",
    "Audio",
]



import asyncio
from datetime import datetime
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any, AsyncIterator

from ..messages.commons import DataEntry

from .base import Node
from .registry import NodeFactory
from ..messages import DataModel, Zero, Message
from ..messages.registry import DataModelFactory

from aiofiles import open
from aiofiles.threadpool.text import AsyncTextIOWrapper
from aiofiles.base import AiofilesContextManager


@NodeFactory.register("record")
class RecordNode(Node[DataModel, Zero]):
    def __init__(
        self,
        record_channel_types: dict[str, str],
        jsonl_file_path: str,
        redis_url: str,
        add_datetime: bool = True,
    ):
        input_channel_types: list[tuple[str, type[DataModel]]] = []
        for channel, channel_type_string in record_channel_types.items():
            input_channel_types.append(
                (channel, DataModelFactory.registry[channel_type_string])
            )
        if add_datetime:
            # add a datetime to jsonl_file_path before the extension. The file can have any extension.
            jsonl_file_path = (
                jsonl_file_path[: jsonl_file_path.rfind(".")]
                + datetime.now().strftime("_%Y-%m-%d_%H-%M-%S")
                + jsonl_file_path[jsonl_file_path.rfind(".") :]
            )

        super().__init__(
            input_channel_types=input_channel_types,
            output_channel_types=[],
            redis_url=redis_url,
        )
        self.jsonl_file_path = jsonl_file_path
        self.aioContextManager: AiofilesContextManager[AsyncTextIOWrapper] | None = None
        self.json_file: AsyncTextIOWrapper | None = None
        self.write_queue: asyncio.Queue[DataEntry[DataModel]] = asyncio.Queue()
        self.write_task: asyncio.Task[None] | None = None

    async def __aenter__(self) -> Self:
        self.aioContextManager = open(self.jsonl_file_path, "w")
        self.json_file = await self.aioContextManager.__aenter__()
        self.write_task = asyncio.create_task(self.write_to_file())
        return await super().__aenter__()

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        if self.aioContextManager:
            await self.aioContextManager.__aexit__(exc_type, exc_value, traceback)
        del self.json_file
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def write_to_file(self) -> None:
        while self.json_file:
            data_entry = await self.write_queue.get()
            await self.json_file.write(data_entry.model_dump_json() + "\n")
            await self.json_file.flush()
            self.write_queue.task_done()

    async def event_handler(
        self, input_channel: str, input_message: Message[DataModel]
    ) -> AsyncIterator[tuple[str, Message[Zero]]]:
        if input_channel in self.input_channel_types:
            await self.write_queue.put(
                DataEntry[self.input_channel_types[input_channel]](  # type: ignore[name-defined]
                    channel=input_channel, data=input_message.data
                )
            )
        else:
            yield input_channel, Message[Zero](data=Zero())



from pydantic import BaseModel, Field

from typing import Generic, Literal, TypeVar


class DataModel(BaseModel):
    data_type: Literal[""] = Field("")


T = TypeVar("T", bound=DataModel)


class Message(BaseModel, Generic[T]):
    data: T = Field(discriminator="data_type")



import base64
from collections import defaultdict
import logging
import toml
from pydantic import BaseModel, ConfigDict, Field

import requests

from ...nodes.registry import NodeFactory


class NodeArgs(BaseModel):
    model_config = ConfigDict(extra="allow")


class NodeConfig(BaseModel):
    node_name: str
    node_class: str
    node_args: NodeArgs = Field(default_factory=NodeArgs)


class Config(BaseModel):
    redis_url: str = Field()
    extra_modules: list[str] = Field(default_factory=lambda: list())
    nodes: list[NodeConfig]


def get_dataflow_config(dataflow_toml: str) -> Config:
    """Get the dataflow configuration from a TOML file.

    Args:
        dataflow_toml (str): Path to the TOML file.

    Returns:
        Config: The dataflow configuration object.

    """
    logger = logging.getLogger(__name__)
    config = Config.model_validate(toml.load(dataflow_toml))
    logger.info(f"Starting dataflow with config {config}")

    return config


def draw_dataflow_mermaid(
    configs: list[Config],
    config_names: list[str] | None = None,
    svg_path: str | None = None,
) -> None:
    """Draw the dataflow graph from the configuration.

    Args:
        config (Config): The dataflow configuration object.

    """

    edge2start_nodes_end_nodes: dict[str, tuple[list[str], list[str]]] = defaultdict(
        lambda: ([], [])
    )
    config_name2nodes: dict[str, list[str]] = defaultdict(list)
    node2config_name: dict[str, str] = {}

    if not config_names:
        config_names = [f"config_{i}" for i in range(len(configs))]
    else:
        assert len(config_names) == len(
            configs
        ), "The number of names should be equal to the number of configs."

    for config, config_name in zip(configs, config_names):
        for module in config.extra_modules:
            __import__(module)

        for node_config in config.nodes:
            node_class = node_config.node_class
            node_name = node_config.node_name
            node_args = node_config.node_args

            node = NodeFactory.make(
                node_class, **node_args.model_dump(), redis_url=config.redis_url
            )
            for input_channel in node.input_channel_types:
                edge2start_nodes_end_nodes[input_channel][1].append(node_name)

            for output_channel in node.output_channel_types:
                edge2start_nodes_end_nodes[output_channel][0].append(node_name)
            node2config_name[node_name] = config_name
            config_name2nodes[config_name].append(node_name)

    graph_str = "flowchart TD\n"
    invisible_nodes: list[str] = []

    for edge, (start_nodes, end_nodes) in edge2start_nodes_end_nodes.items():
        if len(start_nodes) > 1 or len(end_nodes) > 1:
            invisible_edge_node = f"invisible_edge_{edge}"
            invisible_nodes.append(invisible_edge_node)
            if not len(start_nodes):
                hidden_start_node = f"hidden_start_{edge}"
                graph_str += f"    {hidden_start_node}(( )) ---|{edge}| {invisible_edge_node}[ ]\n"
            else:
                for start_node in start_nodes:
                    graph_str += f"    {start_node}[{[start_node]}] ---|{edge}| {invisible_edge_node}[ ]\n"
            if not len(end_nodes):
                hidden_end_node = f"hidden_end_{edge}"
                graph_str += f"    {invisible_edge_node} --- {hidden_end_node}(( ))\n"
            else:
                for end_node in end_nodes:
                    graph_str += (
                        f"    {invisible_edge_node} ---> {end_node}[{[end_node]}]\n"
                    )
            config_name = ""
            if start_nodes:
                config_name = node2config_name[start_nodes[0]]
            elif end_nodes:
                config_name = node2config_name[end_nodes[0]]
            if all(
                [
                    node2config_name[start_node] == config_name
                    for start_node in start_nodes
                ]
            ) and all(
                [node2config_name[end_node] == config_name for end_node in end_nodes]
            ):
                config_name2nodes[config_name].append(invisible_edge_node)
        elif not len(start_nodes):
            hidden_start_node = f"hidden_start_{edge}"
            for end_node in end_nodes:
                graph_str += f"    {hidden_start_node}(( )) --->|{edge}| {end_node}[{[end_node]}]\n"
        elif not len(end_nodes):
            hidden_end_node = f"hidden_end_{edge}"
            for start_node in start_nodes:
                graph_str += f"    {start_node}[{[start_node]}] --->|{edge}| {hidden_end_node}(( ))\n"
        elif len(start_nodes) == 1 and len(end_nodes) == 1:
            graph_str += f"    {start_nodes[0]}[{[start_nodes[0]]}] --->|{edge}| {end_nodes[0]}[{[end_nodes[0]]}]\n"
        else:
            raise ValueError("This should not happen.")

    for config_name, node_names in config_name2nodes.items():
        graph_str += f"subgraph {config_name}\n"
        for node_name in node_names:
            graph_str += f"    {node_name}\n"
        graph_str += "end\n"

    for invisible_node in invisible_nodes:
        graph_str += f"    style {invisible_node} height:0px;\n"

    print(graph_str)

    if svg_path:
        graph_bytes = graph_str.encode("utf8")
        base64_bytes = base64.b64encode(graph_bytes)
        base64_string = base64_bytes.decode("ascii")

        url = f"https://mermaid.ink/svg/{base64_string}"
        response = requests.get(url)
        response.raise_for_status()

        with open(svg_path, "wb") as f:
            f.write(response.content)



from datetime import datetime
from typing import AsyncIterator
from .base import Node
from .registry import NodeFactory

from ..messages import Message, Image, Tick
import os


@NodeFactory.register("performance")
class PerformanceMeasureNode(Node[Tick | Image, Image]):
    def __init__(
        self, input_channel: str, output_channel: str, message_size: int, redis_url: str
    ):
        super().__init__(
            input_channel_types=[
                (input_channel, Tick),
                (output_channel, Image),
            ],
            output_channel_types=[(output_channel, Image)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.message_size = message_size
        self.message_to_datetime: dict[bytes, datetime] = {}

    async def _wait_for_input(self) -> AsyncIterator[tuple[str, Message[Tick | Image]]]:
        async for message in self.pubsub.listen():
            channel = message["channel"].decode("utf-8")
            now = datetime.now()
            if message["type"] == "message" and channel in self.input_channel_types:
                data = Message[self.input_channel_types[channel]].model_validate_json(  # type: ignore
                    message["data"]
                )
                decoded_time = datetime.now()
                if channel == self.output_channel:
                    print(f"Time to decode: {decoded_time - now}")
                yield channel, data
        raise Exception("Input channel closed unexpectedly")

    async def event_loop(
        self,
    ) -> None:
        async for input_channel, input_message in self._wait_for_input():
            async for output_channel, output_message in self.event_handler(
                input_channel, input_message
            ):
                json_str = output_message.model_dump_json()
                self.message_to_datetime[output_message.data.image[:16]] = (
                    datetime.now()
                )
                await self.r.publish(output_channel, json_str)

    async def event_handler(
        self, input_channel: str, input_message: Message[Tick | Image]
    ) -> AsyncIterator[tuple[str, Message[Image]]]:
        match input_message.data:
            case Tick(tick=_):
                size_in_bytes = self.message_size * 1024
                random_bytes = os.urandom(size_in_bytes)
                image = Image(image=random_bytes)
                yield self.output_channel, Message[Image](data=image)
            case Image(image=random_bytes):
                now = datetime.now()
                send_time = self.message_to_datetime[random_bytes[:16]]
                print(f"Latency: {now - send_time}")
                del self.message_to_datetime[random_bytes[:16]]



import asyncio
import sys
from typing import Any, AsyncIterator, Optional, TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..messages.base import DataModel, Message
from .base import Node
from .registry import NodeFactory
from ..messages import Zero, Audio

# For type checking and IDE support
if TYPE_CHECKING:
    import pyaudio

# Runtime import
try:
    import pyaudio
    PYAUDIO_AVAILABLE = True
except ImportError:
    PYAUDIO_AVAILABLE = False

@NodeFactory.register("listener")
class ListenerNode(Node[Zero, Audio]):
    def __init__(self, output_channel: str, redis_url: str):
        if not PYAUDIO_AVAILABLE:
            raise ImportError(
                "PyAudio is not available. Please install it to use the ListenerNode."
            )

        super().__init__(
            input_channel_types=[],
            output_channel_types=[(output_channel, Audio)],
            redis_url=redis_url,
        )
        self.output_channel = output_channel
        self.audio: 'pyaudio.PyAudio' = pyaudio.PyAudio() 
        self.stream: Optional['pyaudio.Stream'] = None
        self.queue: asyncio.Queue[bytes] = asyncio.Queue()
        self.task: Optional[asyncio.Task[None]] = None

    async def __aenter__(self) -> Self:
        if PYAUDIO_AVAILABLE:
            self.stream = self.audio.open(  
                format=pyaudio.paInt16,
                channels=1,
                rate=44100,
                input=True,
                frames_per_buffer=1024,
                stream_callback=self.callback,
            )
            self.task = asyncio.create_task(self.send_frames())
        return await super().__aenter__()

    async def __aexit__(self, _: Any, __: Any, ___: Any) -> None:
        if self.stream:
            self.stream.stop_stream() 
            self.stream.close() 
        if PYAUDIO_AVAILABLE:
            self.audio.terminate() 
        if self.task:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError:
                pass
        return await super().__aexit__(_, __, ___)

    async def send_frames(self) -> None:
        while True:
            frames = await self.queue.get()
            await self.r.publish(
                self.output_channel,
                Message[Audio](data=Audio(audio=frames)).model_dump_json(),
            )

    def callback(
        self, in_data: Optional[bytes], _: Any, __: Any, ___: Any
    ) -> tuple[None, int]:
        if in_data:
            self.queue.put_nowait(in_data)
        return None, pyaudio.paContinue

    async def event_loop(self) -> None:
        while True:
            await asyncio.sleep(1)

    async def event_handler(
        self, _: str, __: Message[Zero]
    ) -> AsyncIterator[tuple[str, Message[Audio]]]:
        raise NotImplementedError("ListenerNode does not have an event handler.")
        yield "", Message[DataModel](data=Zero())


from typing import AsyncIterator


from .base import Node
from .registry import NodeFactory
from ..messages import Tick, Float, Message
import random


@NodeFactory.register("random")
class RandomNode(Node[Tick, Float]):
    def __init__(
        self,
        input_channel: str,
        output_channel: str,
        redis_url: str = "redis://localhost:6379/0",
    ):
        super().__init__(
            input_channel_types=[(input_channel, Tick)],
            output_channel_types=[(output_channel, Float)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel

    async def event_handler(
        self, _: str, __: Message[Tick]
    ) -> AsyncIterator[tuple[str, Message[Float]]]:
        yield self.output_channel, Message[Float](data=Float(value=random.random()))



import asyncio
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import Any, AsyncIterator
from ..messages.base import Message
from ..messages.commons import Audio, Text
from .base import Node
from .registry import NodeFactory
from google.cloud import texttospeech
from google.api_core import exceptions
from google.api_core.client_options import ClientOptions


@NodeFactory.register("tts")
class TTSNode(Node[Text, Audio]):
    def __init__(
        self,
        input_channel: str,
        output_channel: str,
        api_key: str,
        rate: int,
        redis_url: str,
    ) -> None:
        super().__init__(
            input_channel_types=[(input_channel, Text)],
            output_channel_types=[(output_channel, Audio)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.client = texttospeech.TextToSpeechAsyncClient(
            client_options=ClientOptions(api_key=api_key)  # type: ignore[no-untyped-call]
        )
        self.voice = texttospeech.VoiceSelectionParams(
            language_code="en-US", ssml_gender=texttospeech.SsmlVoiceGender.NEUTRAL
        )
        self.audio_config = texttospeech.AudioConfig(
            audio_encoding=texttospeech.AudioEncoding.LINEAR16, sample_rate_hertz=rate
        )
        self.queue = asyncio.Queue[str]()
        self.shutdown = asyncio.Event()

    async def synthesize(self) -> None:
        while not self.shutdown.is_set():
            text = await self.queue.get()
            try:
                input_text = texttospeech.SynthesisInput(text=text)
                response = await self.client.synthesize_speech(
                    request={
                        "input": input_text,
                        "voice": self.voice,
                        "audio_config": self.audio_config,
                    }
                )
                await self.r.publish(
                    self.output_channel,
                    Message[Audio](
                        data=Audio(audio=response.audio_content)
                    ).model_dump_json(),
                )
            except exceptions.GoogleAPICallError as e:
                print(f"Error during speech synthesis: {e}")

    async def __aenter__(self) -> Self:
        self.task = asyncio.create_task(self.synthesize())
        return await super().__aenter__()

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.shutdown.set()
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def event_handler(
        self, input_channel: str, text: Message[Text]
    ) -> AsyncIterator[tuple[str, Message[Audio]]]:
        if input_channel != self.input_channel:
            raise ValueError(f"Unexpected input channel: {input_channel}")
            yield "", Message[Audio](data=Audio(audio=b""))
        else:
            await self.queue.put(text.data.text)



import logging
from typing import Any, Callable, Literal, TypeVar
from pydantic import create_model

from .base import DataModel

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=DataModel)


class DataModelFactory:
    registry: dict[str, type[DataModel]] = {}

    @classmethod
    def register(cls, name: str) -> Callable[[type[T]], type[T]]:
        def inner_wrapper(wrapped_class: type[T]) -> type[T]:
            if name in cls.registry:
                logger.warning("DataModel %s already exists. Will replace it", name)
            new_class = create_model(
                cls.__name__,
                __base__=wrapped_class,
                data_type=(Literal[name], name),
            )
            cls.registry[name] = new_class
            return new_class

        return inner_wrapper

    @classmethod
    def make(cls, name: str, **kwargs: Any) -> DataModel:
        if name not in cls.registry:
            raise ValueError(f"DataModel {name} not found in registry")
        instance: DataModel = cls.registry[name](**kwargs)
        return instance



# import cv2
from .base import DataModel
from .commons import DataEntry

from pydantic import ValidationError

from logging import getLogger


def read_data_from_jsonl(
    jsonl_file_path: str, data_model: type[DataModel], arrange_by_channel: bool = False
) -> list[DataEntry[DataModel]] | dict[str, list[DataEntry[DataModel]]]:
    logger = getLogger(__name__)

    data_entries: list[DataEntry[DataModel]] = []
    data_by_channel: dict[str, list[DataEntry[DataModel]]] = {}

    with open(jsonl_file_path, "r") as f:
        for lineno, line in enumerate(f.readlines()):
            try:
                data_entry = DataEntry[data_model].model_validate_json(line)  # type: ignore[valid-type]
            except ValidationError as e:
                logger.error(
                    f"Validation error at Line {lineno}: {e}. Skipping this line."
                )
                continue
            if arrange_by_channel:
                if data_entry.channel not in data_by_channel:
                    data_by_channel[data_entry.channel] = []
                data_by_channel[data_entry.channel].append(data_entry)
            else:
                data_entries.append(data_entry)

    if arrange_by_channel:
        return data_by_channel
    else:
        return data_entries


# def get_frame_size(frame_bytes: bytes) -> tuple[int, int]:
#     """
#     Get the dimensions (width, height) of a frame in bytes.

#     Args:
#     - frame_bytes: A single frame in bytes.

#     Returns:
#     - (width, height): A tuple representing the frame's dimensions.
#     """
#     # Convert bytes to a numpy array
#     np_frame = np.frombuffer(frame_bytes, dtype=np.uint8)

#     # Decode the numpy array to get the frame
#     frame = cv2.imdecode(np_frame, cv2.IMREAD_COLOR)

#     # Get the dimensions of the frame
#     height, width, _ = frame.shape

#     return (width, height)


# def frames_to_mp4(
#     frame_data_entries: list[DataEntry[Image]], output_mp4_path: str
# ) -> None:
#     """
#     Convert a list of frames in bytes to a video file.

#     Args:
#     - frames: List of frames, where each frame is in bytes.
#     - output_path: Path to save the output video.
#     - fps: Frames per second for the output video.
#     - frame_size: Tuple representing the size of the frames (width, height).

#     Returns:
#     - None
#     """

#     logger = getLogger(__name__)
#     assert len(frame_data_entries) > 0, "Frame data entries list is empty"
#     time_span = (
#         frame_data_entries[-1].timestamp - frame_data_entries[0].timestamp
#     ).total_seconds()
#     fps = len(frame_data_entries) / time_span

#     frames = [frame.data.image for frame in frame_data_entries]
#     frame_size = get_frame_size(frames[0])

#     logger.info(
#         f"Converting {len(frames)} frames to video at {fps} FPS. Frame size: {frame_size}"
#     )

#     # Initialize the video writer
#     fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # type: ignore[attr-defined]
#     out = cv2.VideoWriter(output_mp4_path, fourcc, fps, frame_size)

#     for frame_bytes in frames:
#         # Convert bytes to a numpy array
#         np_frame = np.frombuffer(frame_bytes, dtype=np.uint8)

#         # Decode the numpy array to get the frame
#         frame = cv2.imdecode(np_frame, cv2.IMREAD_COLOR)

#         # Write the frame to the video file
#         out.write(frame)

#     # Release the video writer
#     out.release()



import asyncio
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any, AsyncIterator

from ..messages.base import Message
from ..messages.commons import Audio, Text
from .base import Node
from .registry import NodeFactory

from google.cloud import speech_v1p1beta1 as speech
from google.api_core import exceptions
from google.api_core.client_options import ClientOptions


@NodeFactory.register("transcriber")
class TranscriberNode(Node[Audio, Text]):
    def __init__(
        self,
        input_channel: str,
        output_channel: str,
        rate: int,
        api_key: str,
        redis_url: str,
    ) -> None:
        super().__init__(
            input_channel_types=[(input_channel, Audio)],
            output_channel_types=[(output_channel, Text)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.rate: int = rate
        self.client: speech.SpeechAsyncClient = speech.SpeechAsyncClient(
            client_options=ClientOptions(  # type: ignore[no-untyped-call]
                api_key=api_key
            )
        )
        self.config: speech.RecognitionConfig = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.LINEAR16,
            sample_rate_hertz=self.rate,
            language_code="en-US",
        )
        self.streaming_config: speech.StreamingRecognitionConfig = (
            speech.StreamingRecognitionConfig(config=self.config, interim_results=True)
        )
        self.queue = asyncio.Queue[bytes]()
        self.shutdown = asyncio.Event()

    async def transcribe(
        self,
    ) -> None:
        async def request_generator() -> (
            AsyncIterator[speech.StreamingRecognizeRequest]
        ):
            yield speech.StreamingRecognizeRequest(
                streaming_config=self.streaming_config
            )

            while not self.shutdown.is_set():
                content = await self.queue.get()
                yield speech.StreamingRecognizeRequest(audio_content=content)

        try:
            responses = await self.client.streaming_recognize(
                requests=request_generator()
            )
            async for response in responses:
                if not response.results:
                    continue
                result = response.results[0]
                if not result.alternatives:
                    continue
                transcript = result.alternatives[0].transcript
                if result.is_final:
                    await self.r.publish(
                        self.output_channel,
                        Message[Text](data=Text(text=transcript)).model_dump_json(),
                    )
        except exceptions.GoogleAPICallError as e:
            print(f"Error during transcription: {e}")

    async def __aenter__(self) -> Self:
        self.task = asyncio.create_task(self.transcribe())
        return await super().__aenter__()

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.shutdown.set()
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def event_handler(
        self, input_channel: str, audio: Message[Audio]
    ) -> AsyncIterator[tuple[str, Message[Text]]]:
        if input_channel != self.input_channel:
            raise ValueError(f"Unexpected input channel: {input_channel}")
            yield "", Message[Text](data=Text(text=""))
        else:
            await self.queue.put(audio.data.audio)



import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any, AsyncIterator, Generic, Type, TypeVar
from pydantic import BaseModel, ConfigDict

from abc import abstractmethod
from ..messages import Message
from redis.asyncio import Redis

from ..messages.base import DataModel

InputType = TypeVar("InputType", covariant=True, bound=DataModel)
OutputType = TypeVar("OutputType", covariant=True, bound=DataModel)


class Node(BaseModel, Generic[InputType, OutputType]):
    input_channel_types: dict[str, Type[InputType]]
    output_channel_types: dict[str, Type[OutputType]]
    redis_url: str
    model_config = ConfigDict(extra="allow")

    def __init__(
        self,
        input_channel_types: list[tuple[str, Type[InputType]]],
        output_channel_types: list[tuple[str, Type[OutputType]]],
        redis_url: str = "redis://localhost:6379/0",
    ):
        super().__init__(
            input_channel_types=dict(input_channel_types),
            output_channel_types=dict(output_channel_types),
            redis_url=redis_url,
        )

        self.r: Redis = Redis.from_url(redis_url)
        self.pubsub = self.r.pubsub()

    async def __aenter__(self) -> Self:
        try:
            await self.r.ping()
        except ConnectionError:
            raise ValueError(
                f"Could not connect to Redis with the provided url. {self.redis_url}"
            )
        await self.pubsub.subscribe(*self.input_channel_types.keys())
        return self

    async def __aexit__(self, _: Any, __: Any, ___: Any) -> None:
        await self.pubsub.unsubscribe()
        await self.r.aclose()

    async def _wait_for_input(
        self,
    ) -> AsyncIterator[tuple[str, Message[InputType]]]:
        async for message in self.pubsub.listen():
            channel = message["channel"].decode("utf-8")
            if message["type"] == "message" and channel in self.input_channel_types:
                data = Message[self.input_channel_types[channel]].model_validate_json(  # type: ignore
                    message["data"]
                )
                yield channel, data
        raise Exception("Input channel closed unexpectedly")

    async def event_loop(
        self,
    ) -> None:
        async for input_channel, input_message in self._wait_for_input():
            async for output_channel, output_message in self.event_handler(
                input_channel, input_message
            ):
                await self.r.publish(output_channel, output_message.model_dump_json())
        raise Exception("Event loop exited unexpectedly")

    @abstractmethod
    async def event_handler(
        self, _: str, __: Message[InputType]
    ) -> AsyncIterator[tuple[str, Message[OutputType]]]:
        raise NotImplementedError("event_handler must be implemented in a subclass.")
        yield "", self.output_type()  # unreachable: dummy return value



import asyncio
import logging
import os
import signal
import time
from typing import Annotated, Any, Optional, TypeVar
from ..app import app
from ..reader import get_dataflow_config, draw_dataflow_mermaid, NodeConfig, Config
import typer

from ...nodes import NodeFactory

from subprocess import Popen

import toml
from rq import Queue
from rq.exceptions import InvalidJobOperation
from rq.job import Job
from rq.command import send_stop_job_command
from redis import Redis

InputType = TypeVar("InputType")
OutputType = TypeVar("OutputType")


async def _run_node(node_config: NodeConfig, redis_url: str) -> None:
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger(__name__)
    logger.info(f"Starting node {node_config}")
    try:
        async with NodeFactory.make(
            node_config.node_class,
            **node_config.node_args.model_dump(),
            redis_url=redis_url,
        ) as node:
            logger.info(f"Starting eventloop {node_config.node_name}")
            await node.event_loop()
    except Exception as e:
        logger.error(f"Error in node {node_config.node_name}: {e}")
        raise Exception(e)


def _sync_run_node(node_config: NodeConfig, redis_url: str) -> None:
    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(_run_node(node_config, redis_url))
    except asyncio.CancelledError:
        logger = logging.getLogger(__name__)
        logger.info(f"Node {node_config.node_name} shutdown gracefully.")
    finally:
        loop.close()


@app.command()
def run_node(
    dataflow_toml: str = typer.Option(),
    node_name: str = typer.Option(),
    redis_url: str = typer.Option(),
) -> None:
    logger = logging.getLogger(__name__)
    config = Config.model_validate(toml.load(dataflow_toml))
    logger.info(f"Starting dataflow with config {config}")
    # dynamically import extra modules
    for module in config.extra_modules:
        __import__(module)

    for nodes in config.nodes:
        if nodes.node_name == node_name:
            _sync_run_node(nodes, redis_url)
            break


def _import_extra_modules(extra_modules: list[str]) -> None:
    for module in extra_modules:
        __import__(module)


@app.command()
def run_dataflow(
    dataflow_toml: Annotated[
        str, typer.Argument(help="Configuration dataflow toml file")
    ],
    with_rq: Optional[bool] = typer.Option(
        False, help="Run the dataflow with RQ instead of asyncio."
    ),
) -> None:
    logger = logging.getLogger(__name__)
    config = Config.model_validate(toml.load(dataflow_toml))
    logger.info(f"Starting dataflow with config {config}")

    if with_rq:

        redis = Redis.from_url(config.redis_url)
        queue = Queue(connection=redis)
        job_ids: list[str] = []
        for node in config.nodes:
            job: Job = queue.enqueue(_sync_run_node, node, config.redis_url)
            job_ids.append(job.get_id())

        try:
            # Wait for all jobs to finish
            while not all(Job.fetch(job_id, connection=redis).get_status() == "finished" for job_id in job_ids):
                time.sleep(1)
        except KeyboardInterrupt:
            logger.warning("Terminating RQ jobs.")
            for job_id in job_ids:
                logger.info(f"Terminating job {job_id}")
                try:
                    send_stop_job_command(redis, job_id) # stop the job if it's running
                except InvalidJobOperation:
                    logger.info(f"Job {job_id} is not currently executing. Trying to delete it from queue.")
                job = Job.fetch(job_id, connection=redis)
                job.delete() # remove job from redis
                logger.info(f"Job {job_id} has been terminated. Job status: {job.get_status()}")
        finally:
            return
            

    subprocesses: list[Popen[bytes]] = []

    try:
        # Nodes that run w/ subprocess
        for node in config.nodes:
            command = f"aact run-node --dataflow-toml {dataflow_toml} --node-name {node.node_name} --redis-url {config.redis_url}"
            logger.info(f"executing {command}")
            node_process = Popen(
                [command],
                shell=True,
                preexec_fn=os.setsid,  # Start the subprocess in a new process group
            )
            subprocesses.append(node_process)

        def _cleanup_subprocesses(
            signum: int | None = None, frame: Any | None = None
        ) -> None:
            for node_process in subprocesses:
                try:
                    os.killpg(os.getpgid(node_process.pid), signal.SIGTERM)
                    logger.info(f"Terminating process group {node_process.pid}")
                except ProcessLookupError:
                    logger.warning(
                        f"Process group {node_process.pid} has been terminated."
                    )

        signal.signal(signal.SIGTERM, _cleanup_subprocesses)
        signal.signal(signal.SIGINT, _cleanup_subprocesses)

        for node_process in subprocesses:
            node_process.wait()

    except Exception as e:
        logger.warning("Error in multiprocessing: ", e)
        _cleanup_subprocesses()
    finally:
        _cleanup_subprocesses()


@app.command(help="A nice debugging feature. Draw dataflows with Mermaid.")
def draw_dataflow(
    dataflow_toml: Annotated[
        list[str],
        typer.Argument(
            help="Configuration dataflow toml files. "
            "You can provide multiple toml files to draw multiple dataflows on one graph."
        ),
    ],
    svg_path: Optional[str] = typer.Option(
        None,
        help="Path to save the svg file of the dataflow graph. "
        "If you don't set this, the mermaid code will be printed. "
        "If you set this, we will also render an svg picture of the dataflow graph.",
    ),
) -> None:
    dataflows = [
        get_dataflow_config(single_dataflow_toml)
        for single_dataflow_toml in dataflow_toml
    ]

    draw_dataflow_mermaid(dataflows, dataflow_toml, svg_path)



from .base import Node
from .tick import TickNode
from .random import RandomNode
from .record import RecordNode
from .listener import ListenerNode
from .speaker import SpeakerNode
from .transcriber import TranscriberNode
from .performance import PerformanceMeasureNode
from .print import PrintNode
from .tts import TTSNode
from .registry import NodeFactory

__all__ = [
    "Node",
    "TickNode",
    "RandomNode",
    "NodeFactory",
    "RecordNode",
    "ListenerNode",
    "SpeakerNode",
    "TranscriberNode",
    "PerformanceMeasureNode",
    "PrintNode",
    "TTSNode",
]



import logging
from typing import Any, Callable, TypeVar

from ..messages.base import DataModel
from .base import Node

logger = logging.getLogger(__name__)
InputType = TypeVar("InputType", bound=DataModel)
OutputType = TypeVar("OutputType", bound=DataModel)


class NodeFactory:
    registry: dict[str, type[Node[DataModel, DataModel]]] = {}

    @classmethod
    def register(
        cls, name: str
    ) -> Callable[
        [type[Node[InputType, OutputType]]], type[Node[InputType, OutputType]]
    ]:
        def inner_wrapper(
            wrapped_class: type[Node[InputType, OutputType]],
        ) -> type[Node[InputType, OutputType]]:
            if name in cls.registry:
                logger.warning("Executor %s already exists. Will replace it", name)
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def make(cls, name: str, **kwargs: Any) -> Node[DataModel, DataModel]:
        if name not in cls.registry:
            raise ValueError(f"Executor {name} not found in registry")
        return cls.registry[name](**kwargs)
